﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Problems.VehicleRouting.Encodings;
using HeuristicLab.Problems.VehicleRouting.Variants;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Data;
using HeuristicLab.Common;
using HeuristicLab.Parameters;
using HeuristicLab.Problems.VehicleRouting.Encodings.Potvin;
using HeuristicLab.Problems.VehicleRouting;
using HeuristicLab.Problems.VehicleRouting.Interfaces;
using HeuristicLab.Optimization;

namespace HeuristicLab.PDPSimulation.Operators.Improvers {
  [Item("IntraRouteImprovementOperator", "An operator that improves vehicle routing solutions.")]
  [StorableClass]
  public class IntraRouteImprovementOperator : PotvinManipulator {
    public ValueLookupParameter<ResultCollection> ResultsParameter {
      get { return (ValueLookupParameter<ResultCollection>)Parameters["Results"]; }
    }

    public IValueParameter<IntValue> ExecutionIntervalParameter {
      get { return (IValueParameter<IntValue>)Parameters["ExecutionInterval"]; }
    }

    public IValueParameter<IntValue> ImprovementAttemptsParameter {
      get { return (IValueParameter<IntValue>)Parameters["ImprovementAttempts"]; }
    }

    public IValueParameter<IntValue> SampleSizeParameter {
      get { return (IValueParameter<IntValue>)Parameters["SampleSize"]; }
    }

    [StorableConstructor]
    private IntraRouteImprovementOperator(bool deserializing) : base(deserializing) { }
    private IntraRouteImprovementOperator(IntraRouteImprovementOperator original, Cloner cloner) : base(original, cloner) { }
    public IntraRouteImprovementOperator()
      : base() {
      Parameters.Add(new ValueParameter<IntValue>("ExecutionInterval", "The execution interval.", new IntValue(10)));
      Parameters.Add(new ValueParameter<IntValue>("ImprovementAttempts", "The number of iterations that should be executed.", new IntValue(100)));
      Parameters.Add(new ValueParameter<IntValue>("SampleSize", "The number of moves that should be sampled per iteration.", new IntValue(1000)));
      Parameters.Add(new ValueLookupParameter<ResultCollection>("Results", "The result collection where the best VRP solution should be stored."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new IntraRouteImprovementOperator(this, cloner);
    }

    protected override void Manipulate(IRandom random, PotvinEncoding individual) {
      ResultCollection results = ResultsParameter.ActualValue;
      int generations = -1;
      if (results.ContainsKey("Iterations"))
        generations = (results["Iterations"].Value as IntValue).Value;
      else if (results.ContainsKey("Generations"))
        generations = (results["Generations"].Value as IntValue).Value;

      int interval = ExecutionIntervalParameter.Value.Value;

      if (generations % interval == 0) {
        int attempts = ImprovementAttemptsParameter.Value.Value;
        int sampleSize = SampleSizeParameter.Value.Value;
        IVRPProblemInstance instance = ProblemInstance;
        IPickupAndDeliveryProblemInstance pdp = instance as IPickupAndDeliveryProblemInstance;

        double quality = instance.Evaluate(individual).Quality;

        for (int i = 0; i < attempts; i++) {
          for (int sample = 0; sample < sampleSize; sample++) {
            int tourIdx = random.Next(individual.Tours.Count);
            Tour tour = individual.Tours[tourIdx];
            ItemList<Tour> original = individual.Tours.Clone() as ItemList<Tour>;

            if (tour.Stops.Count > 1) {
              int idx1 = random.Next(tour.Stops.Count);
              int idx2 = random.Next(tour.Stops.Count - 1);
              if (idx2 >= idx1)
                idx2++;

              int city1 = tour.Stops[idx1];
              int city2 = tour.Stops[idx2];

              tour.Stops[idx1] = city2;
              tour.Stops[idx2] = city1;

              if (pdp != null && random.NextDouble() < 0.5) {
                int target1 = -1;
                int targetIdx1 = -1;
                int target2 = -1;
                int targetIdx2 = -1;

                target1 = pdp.GetPickupDeliveryLocation(city1);
                if (target1 == city1 || target1 <= 0)
                  targetIdx1 = -1;
                else {
                  targetIdx1 = tour.Stops.IndexOf(target1);

                  if (targetIdx1 == -1) {
                    individual.Tours.First(t => t.Stops.Contains(target1)).Stops.Remove(target1);
                    int pos = individual.FindBestInsertionPlace(tour, target1);
                    tour.Stops.Insert(pos, target1);
                    targetIdx1 = pos;
                  }
                }

                target2 = pdp.GetPickupDeliveryLocation(city2);
                if (target2 == city2 || target2 <= 0)
                  targetIdx2 = -1;
                else {
                  targetIdx2 = tour.Stops.IndexOf(target2);

                  if (targetIdx2 == -1) {
                    individual.Tours.First(t => t.Stops.Contains(target2)).Stops.Remove(target2);
                    int pos = individual.FindBestInsertionPlace(tour, target2);
                    tour.Stops.Insert(pos, target2);
                    targetIdx2 = pos;

                    if (targetIdx1 >= pos)
                      targetIdx1++;
                  }
                }

                if (targetIdx1 != -1 && targetIdx2 != -1) {
                  tour.Stops[targetIdx1] = target2;
                  tour.Stops[targetIdx2] = target1;
                } 
              }

              double newQuality = instance.Evaluate(individual).Quality;
              if (newQuality < quality) {
                quality = newQuality;
                break;
              } else {
                //revert
                individual.Tours = original;
              }
            }
          }
        }

        IAtomicOperation op = this.ExecutionContext.CreateChildOperation(instance.SolutionEvaluator);
        op.Operator.Execute((IExecutionContext)op, CancellationToken);
      }
    }
  }
}
